import { GraphQLError } from 'graphql';
import { IResolvers } from 'mercurius';
import { withAuth } from '../utils';
import { MutationResolvers } from '../generated/graphql';

export const ROOM_ADDED = 'ROOM_ADDED';
export const ROOM_UPDATED = 'ROOM_UPDATED';

export const roomResolvers: IResolvers = {
  Query: {
    rooms: async (_, __, { prisma }) => {
      return prisma.room.findMany({
        where: { isPrivate: false },
      });
    },
    room: async (_, { id }, { prisma }) => {
      return prisma.room.findUnique({
        where: { id },
      });
    },
  },
  Mutation: {
    createRoom: withAuth<MutationResolvers['createRoom']>(
      async (_, { name, description, isPrivate }, { prisma, jwt, pubsub }) => {
        const room = await prisma.room.create({
          data: {
            name,
            description,
            isPrivate: isPrivate ?? false,
            owner: {
              connect: { id: jwt.sub },
            },
            members: {
              connect: { id: jwt.sub },
            },
          },
        });

        pubsub.publish({
          topic: ROOM_ADDED,
          payload: { roomAdded: room },
        });

        return room;
      }
    ),
    joinRoom: withAuth<MutationResolvers['joinRoom']>(
      async (_, { roomId }, { prisma, jwt, pubsub }) => {
        const room = await prisma.room.findUnique({
          where: { id: roomId },
          include: { members: true },
        });

        if (!room) {
          throw new GraphQLError('Room not found', {
            extensions: {
              code: 'NOT_FOUND',
            },
          });
        }

        if (room.isPrivate) {
          // In a real application, you would check if the user has been invited
          throw new GraphQLError(
            'You cannot join a private room without an invitation',
            {
              extensions: {
                code: 'FORBIDDEN',
              },
            }
          );
        }

        // Check if user is already a member
        const isMember = room.members.some(
          (member: { id: string }) => member.id === jwt.sub
        );
        if (isMember) {
          return room;
        }

        const updatedRoom = await prisma.room.update({
          where: { id: roomId },
          data: {
            members: {
              connect: { id: jwt.sub },
            },
          },
          include: { members: true },
        });

        // Publish room updated event
        pubsub.publish({
          topic: ROOM_UPDATED,
          payload: { roomUpdated: updatedRoom },
        });

        return updatedRoom;
      }
    ),
    leaveRoom: withAuth<MutationResolvers['leaveRoom']>(
      async (_, { roomId }, { prisma, jwt, pubsub }) => {
        const room = await prisma.room.findUnique({
          where: { id: roomId, members: { some: { id: jwt.sub } } },
          include: { members: true },
        });

        if (!room) {
          throw new GraphQLError('Room not found', {
            extensions: {
              code: 'NOT_FOUND',
            },
          });
        }

        if (room.ownerId === jwt.sub) {
          throw new GraphQLError('You cannot leave a room you own', {
            extensions: {
              code: 'FORBIDDEN',
            },
          });
        }

        const updatedRoom = await prisma.room.update({
          where: { id: roomId },
          data: {
            members: {
              disconnect: { id: jwt.sub },
            },
          },
          include: { members: true },
        });

        // Publish room updated event
        pubsub.publish({
          topic: ROOM_UPDATED,
          payload: { roomUpdated: updatedRoom },
        });

        return true;
      }
    ),
  },
  Subscription: {
    roomAdded: {
      subscribe: (_, __, { pubsub }) => pubsub.subscribe(ROOM_ADDED),
    },
    roomUpdated: {
      subscribe: (_, __, { pubsub }) => pubsub.subscribe(ROOM_UPDATED),
    },
  },
  Room: {
    messages: async (parent: any, _, { prisma }) =>
      parent.messages
        ? parent.messages
        : prisma.room
            .findUnique({
              where: { id: parent.id },
            })
            .messages({
              orderBy: { createdAt: 'asc' },
            }),
    members: async (parent, _, { prisma }) =>
      parent.members
        ? parent.members
        : prisma.room
            .findUnique({
              where: { id: parent.id },
            })
            .members(),
    owner: async (parent, _, { prisma }) =>
      parent.owner
        ? parent.owner
        : prisma.room
            .findUnique({
              where: { id: parent.id },
            })
            .owner(),
  },
};
